import os

import pytest
import yaml

from distgen.multispec import Multispec, MultispecError


here = os.path.dirname(__file__)
fixtures = os.path.join(here, 'fixtures')
ms_fixtures = os.path.join(fixtures, 'multispec')
simplest = {'version': '1',
            'specs': {'distroinfo': {'fedora': {'distros': ['fedora-26-x86_64']}}}
            }


class TestMultispec(object):
    @pytest.mark.parametrize('proj_dir, path', [
        (ms_fixtures, 'simplest.yaml'),
        (fixtures, os.path.join('multispec', 'simplest.yaml')),
    ])
    def test_from_path_ok(self, proj_dir, path):
        assert Multispec.from_path(proj_dir, path).raw_data == simplest

    def test_from_path_nok(self):
        with pytest.raises(SystemExit):
            Multispec.from_path('.', 'nope')

    def test_validate(self):
        with open(os.path.join(ms_fixtures, 'complex.yaml')) as f:
            Multispec(yaml.load(f))._validate()

    def test_has_spec_group(self):
        ms = Multispec.from_path(ms_fixtures, 'complex.yaml')
        assert ms.has_spec_group('something_else')
        assert not ms.has_spec_group('nope')

    def test_get_spec_group(self):
        ms = Multispec.from_path(ms_fixtures, 'complex.yaml')
        assert ms.get_spec_group('version') == \
            {'2.2': {'version': '2.2'}, '2.4': {'version': '2.4'}}
        with pytest.raises(KeyError):
            ms.get_spec_group('nope')

    def test_has_spec_group_item(self):
        ms = Multispec.from_path(ms_fixtures, 'complex.yaml')
        assert ms.has_spec_group_item('something_else', 'foo')
        assert not ms.has_spec_group_item('something_else', 'nope')

    def test_get_spec_group_item(self):
        ms = Multispec.from_path(ms_fixtures, 'complex.yaml')
        assert ms.get_spec_group_item('something_else', 'foo') == {'spam': 'ham'}
        with pytest.raises(KeyError):
            ms.get_spec_group_item('something_else', 'nope')

    def test_get_all_combinations_simple(self):
        ms = Multispec.from_path(ms_fixtures, 'simplest.yaml')
        assert list(ms.get_all_combinations()) == [{'distro': 'fedora-26-x86_64'}]

    def test_get_all_combinations_complex(self):
        ms = Multispec.from_path(ms_fixtures, 'complex.yaml')
        def comb_key(c):
            return (c['distro'], c['something_else'], c['version'])
        assert sorted(ms.get_all_combinations(), key=comb_key) == sorted([
            {'distro': 'centos-7-x86_64', 'something_else': 'bar', 'version': '2.2'},
            {'distro': 'centos-7-x86_64', 'something_else': 'bar', 'version': '2.4'},
            {'distro': 'centos-7-x86_64', 'something_else': 'foo', 'version': '2.2'},
            {'distro': 'centos-7-x86_64', 'something_else': 'foo', 'version': '2.4'},
            {'distro': 'centos-7-x86_64', 'something_else': 'baz', 'version': '2.2'},
            {'distro': 'centos-7-x86_64', 'something_else': 'baz', 'version': '2.4'},
            {'distro': 'fedora-26-x86_64', 'something_else': 'bar', 'version': '2.2'},
            {'distro': 'fedora-26-x86_64', 'something_else': 'bar', 'version': '2.4'},
            {'distro': 'fedora-26-x86_64', 'something_else': 'foo', 'version': '2.4'},
            {'distro': 'fedora-26-x86_64', 'something_else': 'baz', 'version': '2.2'},
            {'distro': 'fedora-26-x86_64', 'something_else': 'baz', 'version': '2.4'},
            {'distro': 'fedora-25-x86_64', 'something_else': 'bar', 'version': '2.2'},
            {'distro': 'fedora-25-x86_64', 'something_else': 'foo', 'version': '2.2'},
            {'distro': 'fedora-25-x86_64', 'something_else': 'foo', 'version': '2.4'},
            {'distro': 'fedora-25-x86_64', 'something_else': 'baz', 'version': '2.2'},
            {'distro': 'fedora-25-x86_64', 'something_else': 'baz', 'version': '2.4'},
        ], key=comb_key)

    def test_parse_selectors_ok(self):
        ms = Multispec.from_path(ms_fixtures, 'simplest.yaml')
        assert ms.parse_selectors(['foo=bar', 'spam=ham']) == {'foo': 'bar', 'spam': 'ham'}

    def test_parse_selectors_nok(self):
        ms = Multispec.from_path(ms_fixtures, 'simplest.yaml')
        with pytest.raises(MultispecError):
            ms.parse_selectors(['foo bar'])

    def test_normalize_distro(self):
        ms = Multispec.from_path(ms_fixtures, 'simplest.yaml')
        assert ms.normalize_distro('foo/bar/fedora-26-x86_64.yaml') == 'fedora-26-x86_64'

    def test_verify_selectors_ok(self):
        ms = Multispec.from_path(ms_fixtures, 'complex.yaml')
        assert ms.verify_selectors(['version=2.4', 'something_else=foo'],
                                   distro='fedora-26-x86_64') == (True, '')

    @pytest.mark.parametrize('selectors, distro, msg', [
        (['distroinfo=fedora'], 'fedora-26-x86_64', '"distroinfo" not allowed in selectors, it is '
         'chosen automatically based on distro'),
        (['version=2.2'], 'fedora-26-x86_64', '"something_else" selector must be present'),
        (['xxx=asd'], 'fedora-26-x86_64', '"xxx" not an entry in specs'),
        (['version=3.4'], 'fedora-26-x86_64', '"3.4" not an entry in specs.version'),
        (['version=2.2', 'something_else=foo'], 'fedora-26-x86_64', 'This combination is excluded '
         'in matrix section'),
        (['version=2.4', 'something_else=foo'], 'fedora-27-x86_64', '"fedora-27-x86_64" distro '
         'not found in any specs.distroinfo.*.distros section'),
    ])
    def test_verify_selectors_nok(self, selectors, distro, msg):
        ms = Multispec.from_path(ms_fixtures, 'complex.yaml')
        assert ms.verify_selectors(selectors, distro) == (False, msg)

    def test_select_data_ok(self):
        ms = Multispec.from_path(ms_fixtures, 'complex.yaml')
        assert ms.select_data(['version=2.4', 'something_else=foo'], 'fedora-26-x86_64') == \
            {'authoritative_source_url': 'some.url.fedoraproject.org',
             'distro_specific_help': 'Some Fedora specific help',
             'spam': 'ham',
             'vendor': 'Fedora Project',
             'version': '2.4'}

    def test_select_data_nok(self):
        ms = Multispec.from_path(ms_fixtures, 'complex.yaml')
        with pytest.raises(MultispecError):
            ms.select_data(['version=2.4', 'something_else=foobar'], 'fedora-27-x86_64')



import os

from distgen.config import merge_yaml, load_config


class TestMergeYaml(object):
    def test_non_overlapping(self):
        a = {'a': 1, 'aa': 2}
        b = {'b': 3, 'bb': 4}
        assert merge_yaml(a, b) == {'a': 1, 'aa': 2, 'b': 3, 'bb': 4}

    def test_overlapping(self):
        a = {'a': 1, 'aa': 2, 'c': 111}
        b = {'b': 3, 'bb': 4, 'c': 112}
        assert merge_yaml(a, b) == {'a': 1, 'aa': 2, 'c': 112, 'b': 3, 'bb': 4}

    def test_nested(self):
        a = {'a': {'b': 'c', 'd': 'e'}}
        b = {'a': {'b': 'x', 'f': 'g'}}
        assert merge_yaml(a, b) == {'a': {'b': 'x', 'd': 'e', 'f': 'g'}}


class TestLoadConfig(object):
    p = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'fixtures', 'load_config')

    def test_simple(self):
        assert load_config([self.p], 'simple.yaml') == {'foo': 'bar', 'ham': {'and': 'spam'}}

    def test_recursive(self):
        assert load_config([self.p], 'level3.yaml') == \
            {'extends': 'level2.yaml', 'foo': 'override', 'ham': {'and': 'spam', 'or': 'eggs'}}



import os

import pytest

from distgen.pathmanager import PathManager


this = os.path.abspath(__file__)
this_fn = os.path.basename(this)
here = os.path.dirname(this)
doesnt_exist = this + '.hahahaha'
open_file_type = type(open(this))


class TestPathmanager(object):
    def test_get_path(self):
        pm = PathManager(['some', 'path'])
        assert pm.get_path() == [os.getcwd()] + ['some', 'path']

        pm = PathManager(['some', 'path'], 'SOME_PATH_VAR')
        os.environ['SOME_PATH_VAR'] = 'foo:bar'
        assert pm.get_path() == [os.getcwd()] + ['foo', 'bar', 'some', 'path']

    def test_get_file_abspath(self):
        pm = PathManager([])
        assert pm.get_file(this) == this
        assert pm.get_file(doesnt_exist) is None

    @pytest.mark.parametrize('path, prefered_path, fail, retval', [
        ([], None, False, None),
        ([here], None, False, this),
        (['some/other/path', here], None, False, this),
        (['some/other/path'], [here], False, this),
        ([], None, True, SystemExit(1)),
    ])
    def test_get_file_relpath(self, path, prefered_path, fail, retval):
        pm = PathManager(path)
        if isinstance(retval, SystemExit):
            with pytest.raises(SystemExit) as e:
                pm.get_file(this_fn, prefered_path=prefered_path, fail=fail)
            assert e.value.code == retval.code
        else:
            assert pm.get_file(this_fn, prefered_path=prefered_path, fail=fail) == retval

    @pytest.mark.parametrize('path, prefered_path, fail, retval', [
        ([], None, False, type(None)),
        ([here], None, False, open_file_type),
        (['some/other/path', here], None, False, open_file_type),
        (['some/other/path'], [here], False, open_file_type),
        ([], None, True, SystemExit(1)),
    ])
    def test_get_file_relpath(self, path, prefered_path, fail, retval):
        pm = PathManager(path)
        if isinstance(retval, SystemExit):
            with pytest.raises(SystemExit) as e:
                pm.open_file(this_fn, prefered_path=prefered_path, fail=fail)
            assert e.value.code == retval.code
        else:
            assert isinstance(pm.open_file(this_fn, prefered_path=prefered_path, fail=fail),
                              retval)



from __future__ import print_function
import six

import sys
import copy

from distgen.err import fatal
from distgen.pathmanager import PathManager


def _merge_yaml(origin, override):
    """
    Merge simple yaml node recursively.  If the node is non-dict, return itself,
    otherwise recurse down for each item.
    """
    if isinstance(origin, dict) and isinstance(override, dict):
        for k, v in six.iteritems(override):
            if k in origin:
                origin[k] = _merge_yaml(origin[k], override[k])
            else:
                origin[k] = copy.deepcopy(override[k])
        return origin

    return copy.deepcopy(override)


def merge_yaml(origin, override):
    old = copy.deepcopy(origin)
    new = copy.deepcopy(override)
    return _merge_yaml(old, new)


def __recursive_load(pm, stack, filename):
    if filename in stack:
        fatal("already parsed " + filename)

    stack.append(filename)

    import yaml
    try:
        yaml_data = yaml.load(pm.open_file(
            filename,
            fail=True,
            file_desc="configuration file",
        ))
    except yaml.YAMLError as exc:
        print("Error in configuration file: {0}".format(exc))
        sys.exit(1)

    if yaml_data and "extends" in yaml_data:
        subdata = __recursive_load(pm, stack, yaml_data["extends"])
        yaml_data = merge_yaml(subdata, yaml_data)

    return yaml_data


def load_config(path, config_file, context=None):
    yaml_data = __recursive_load(PathManager(path), [], config_file)
    return yaml_data



config = {
    'macros': {
        # Directory under which 'atomic CMD' will mount Host's '/' directory
        # into (privileged) container.
        'atomic_hostdir': '/host',

        # Basic "privileged" docker command
        'atomic_docker_pcmd':
            "docker run -t -i --rm --privileged -u 0:0 "  + \
            "-v /:$atomic_hostdir --net=host --ipc=host --pid=host " + \
            "-e HOST=$atomic_hostdir " + \
            '-e LOGDIR=/var/log/"\\${NAME}" ' + \
            '-e DATADIR=/var/lib/"\\${NAME}" ' + \
            '-e CONFDIR=/etc/"\\${NAME}" ' + \
            '-e IMAGE="\\${IMAGE}" -e NAME="\\${NAME}" ' + \
            '-e OPT1 -e OPT2 -e OPT3 ' + \
            '\\${OPT2} \\${IMAGE}',

        # Atomic commands, should be redefined in dependant project.
        'atomic_install': '$atomic_docker_pcmd install',
        'atomic_uninstall': '$atomic_docker_pcmd install',
    }
}



from __future__ import print_function
import os, sys


class PathManager(object):
    envvar = None

    def __init__(self, path, envvar=None):
        self.path = path
        self.envvar = envvar


    def get_file(self, filename, prefered_path=None, fail=False,
                 file_desc="file"):

        if filename.startswith('/'):
            if os.path.isfile(filename):
                return filename
            else:
                return None

        path = self.get_path()
        if prefered_path:
            path = prefered_path + path

        for i in path:
            config_file = i + "/" + filename
            if os.path.isfile(config_file):
                return config_file

        if fail:
            print("can't find {0} '{1}'".format(file_desc, filename))
            sys.exit(1)

        return None


    def open_file(self, relative, prefered_path=None,
                  fail=False, file_desc="file"):

        filename = self.get_file(relative, prefered_path, fail=fail,
                                 file_desc=file_desc)
        if not filename:
            return None

        try:
            fd = open(filename)
        except IOError as err:
            if fail:
                print("can't open file {0}".format(relative))
                sys.exit(1)
            return None

        return fd


    def get_path(self):
        path = self.path
        if self.envvar and self.envvar in os.environ:
            env_path = os.environ[self.envvar].split(':')
            path = env_path + path

        return [os.getcwd()] + path



import copy
import itertools
import os

import yaml

from distgen.config import merge_yaml
from distgen.pathmanager import PathManager


DISTROINFO_GRP = 'distroinfo'
DISTROINFO_GRP_DISTROS = 'distros'


class MultispecError(Exception):
    def __init__(self, cause):
        self.cause = cause

    def __str__(self):
        return self.cause


class Multispec(object):
    def __init__(self, data):
        self.raw_data = data
        self._validate()
        self._version = None
        self._specgroups = {}
        self._matrix = {}
        self._process()

    @classmethod
    def from_path(cls, project_dir, path):
        pm = PathManager([])
        fd = pm.open_file(path, [project_dir], fail=True)
        data = yaml.load(fd)
        return cls(data)

    def _process(self):
        self._version = int(self.raw_data['version'])
        self._specgroups = self.raw_data['specs']
        self._matrix = self.raw_data.get('matrix', {})  # optional

    def _validation_err(self, err):
        raise MultispecError('Multispec validation error: ' + err)

    def _validate(self):
        # TODO: perhaps use jsonschema or something similar?
        if not isinstance(self.raw_data, dict):
            self._validation_err('multispec must be a mapping, is "{0}"'.
                                 format(type(self.raw_data)))
        self._validate_version(self.raw_data.get('version', None))
        self._validate_specs(self.raw_data.get('specs', None))
        self._validate_matrix(self.raw_data.get('matrix', {}))

    def _validate_version(self, version):
        try:
            v = int(version)
        except TypeError:
            self._validation_err('version must be int or string convertable to string, is "{0}"'.
                                 format(version))
        if v != 1:
            self._validation_err('only version "1" is recognized by this distgen version')

    def _validate_specs(self, specs):
        if not isinstance(specs, dict):
            self._validation_err('"specs" must be a mapping, is "{0}"'.format(type(specs)))

        for k, v in specs.items():
            self._validate_spec_group(k, v)

        if 'distroinfo' not in specs:
            self._validation_err('"distroinfo" must be in "specs"')

    def _validate_spec_group(self, name, spec_group):
        if not isinstance(spec_group, dict):
            self._validation_err('a spec group "{0}" must be a mapping, is "{1}"'.
                                 format(name, type(spec_group)))
        for k, v in spec_group.items():
            self._validate_single_spec(name, k, v)

    def _validate_single_spec(self, groupname, specname, spec):
        if not isinstance(spec, dict):
            self._validation_err('a spec "{0}" must be a mapping, is "{1}"'.
                                 format(specname, type(spec)))

        if groupname == 'distroinfo':
            if 'distros' not in spec:
                self._validation_err('"distroinfo" spec "{0}" must contain "distros" list'.
                                    format(specname))
            self._validate_distros(spec['distros'])

    def _validate_distros(self, distros):
        if not isinstance(distros, list):
            self._validation_err('"distros" entry must be a list, is "{0}"'.format(type(distros)))

    def _validate_matrix(self, matrix):
        if not isinstance(matrix, dict):
            self._validation_err('matrix must be a mapping, is "{0}"'.format(type(matrix)))

        self._validate_excludes(matrix.get('exclude', []))

    def _validate_excludes(self, excludes):
        if not isinstance(excludes, list):
            self._validation_err('matrix.exclude must be a list, is "{0}"'.format(type(excludes)))

        for e in excludes:
            self._validate_single_exclude(e)

    def _validate_single_exclude(self, exclude):
        if not isinstance(exclude, dict):
            self._validation_err('each exclude must be a mapping, not "{0}"'.format(type(exclude)))

        if 'distroinfo' in exclude:
            self._validation_err('a matrix.exclude member must not contain "distroinfo", '
                                 'use "distro" list instead')

        if 'distros' in exclude and not isinstance(exclude['distros'], list):
            self._validation_err('matrix.exclude.*.distros must be a list, found "{0}"'.
                                 format(type(exclude['distros'])))

    def has_spec_group(self, group):
        return group in self._specgroups

    def get_spec_group(self, group):
        return copy.deepcopy(self._specgroups[group])

    def get_distroinfos_by_distro(self, distro):
        distro = self.normalize_distro(distro)
        distroinfos = []

        for di in self.get_spec_group(DISTROINFO_GRP):
            item = self.get_spec_group_item(DISTROINFO_GRP, di)
            if distro in item[DISTROINFO_GRP_DISTROS]:
                distroinfos.append(item)

        return distroinfos

    def has_spec_group_item(self, group, item):
        return self.has_spec_group(group) and item in self.get_spec_group(group)

    def get_spec_group_item(self, group, item):
        return copy.deepcopy(self.get_spec_group(group)[item])

    def get_all_combinations(self):
        distros = set()

        for spec_name, spec_vals in self.get_spec_group(DISTROINFO_GRP).items():
            distros.update(spec_vals[DISTROINFO_GRP_DISTROS])

        values = [[{'group': 'distro', 'value': d} for d in distros]]
        for grp_name, grp_vals in self._specgroups.items():
            if grp_name != DISTROINFO_GRP:
                values.append([{'group': grp_name, 'value': val} for val in grp_vals.keys()])

        for combination in itertools.product(*values):
            selector_dict = {p['group']: p['value'] for p in combination}
            as_selectors = ['{0}={1}'.format(k, v)
                            for k, v in selector_dict.items() if k != 'distro']
            if self.verify_selectors(as_selectors, selector_dict['distro'])[0]:
                yield selector_dict

    def verify_selectors(self, selectors, distro):
        parsed_selectors = self.parse_selectors(selectors)
        distro = self.normalize_distro(distro)

        if DISTROINFO_GRP in parsed_selectors.keys():
            return False, \
                '"{0}" not allowed in selectors, it is chosen automatically based on distro'.\
                format(DISTROINFO_GRP)

        # first, verify that these selector values even exist in specs
        for selector_name, selector_val in parsed_selectors.items():
            if not self.has_spec_group(selector_name):
                return False, '"{0}" not an entry in specs'.format(selector_name)
            elif not self.has_spec_group_item(selector_name, selector_val):
                return False, '"{0}" not an entry in specs.{1}'.format(
                    selector_val, selector_name)

        for selector_name in self.raw_data['specs'].keys():
            if selector_name != DISTROINFO_GRP and selector_name not in parsed_selectors:
                return False, '"{0}" selector must be present'.format(selector_name)

        # second, verify that these selector values are not excluded by matrix
        for excluded in self._matrix.get('exclude', []):
            exclude = True
            for k, v in excluded.items():
                if k == DISTROINFO_GRP_DISTROS and distro not in v:
                    exclude = False
                elif k != DISTROINFO_GRP_DISTROS and parsed_selectors[k] != v:
                    exclude = False
            if exclude:
                return False, 'This combination is excluded in matrix section'

        # third, make sure we have a distroinfo section that contains passed distro
        if not self.get_distroinfos_by_distro(distro):
            return False, '"{0}" distro not found in any specs.distroinfo.*.distros section'.\
                format(distro)

        return True, ''

    def select_data(self, selectors, distro):
        allowed, reason = self.verify_selectors(selectors, distro)
        if not allowed:
            raise MultispecError(reason)

        parsed_selectors = self.parse_selectors(selectors)
        selected_data = {}

        for selector_name, selector_val in parsed_selectors.items():
            spec_content = self.get_spec_group_item(selector_name, selector_val)
            selected_data = merge_yaml(selected_data, spec_content)

        for di in self.get_distroinfos_by_distro(distro):
            # we don't want the list of allowed distros to end up in spec
            di = copy.deepcopy(di)
            di.pop(DISTROINFO_GRP_DISTROS)
            selected_data = merge_yaml(selected_data, di)

        return selected_data

    def parse_selectors(self, selectors):
        res = []
        for selector in selectors:
            if selector.count('=') != 1:
                raise MultispecError('selector "{0}" does not contain exactly one "="'.
                                     format(selector))
            else:
                res.append(selector.split('='))
        return dict(res)

    def normalize_distro(self, distro):
        return os.path.splitext(os.path.basename(distro))[0]



#!/bin/python

from distgen.project import AbstractProject

class Project(AbstractProject):

    maintainer = "Pavel Raiskup <praiskup@redhat.com>"

    def inst_init(self, spec, template, sysconf):
        """
        Demonstrate logic in inst_init().
        """
        test_var = "{0}-{1}-{2}".format(sysconf['os']['id'],
                                        sysconf['os']['version'],
                                        sysconf['os']['arch'])

        self.test_var = "build_for {0}".format(test_var)


    def inst_finish(self, spec, tpl, config, data):

        # We can in-place update 'data' variable generated from yaml
        # specification.
        if config['os']['id'] == 'fedora':
            action = {'type': 'pkg'}
            action['action'] = "install"
            action['packages'] = ['fedpkg']

            data['parts']['pkginstall']['data'].append(action)



config = {
    'name': 'remote-pkg',
    'macros': {
        'dont_replace': '$pkgdatadir/blah',
        'replace_by_second_project': 'not-replaced',
        'to_be_replaced': 'value-to-be-replaced',
    }
}



"""
This file contains helper classes for jinja templating system, because its much
more comfortable and secure to write command-class, than implement that
completely in template.
"""

class Command(object):
    interactive = False

    def __init__(self, config):
        self.config = config


    def is_interactive(self, opts=None):
        i = self.config.interactive
        if opts and 'interactive' in opts:
            i = opts['interactive']
        return i


    def in_container(self, opts=None):
        return self.config.container


class AbstractPkgManger(Command):
    pass


class YumPkgManager(AbstractPkgManger):
    binary = "yum"

    def _base_command(self, opts):
        base = self.binary + " -y"
        if self.is_interactive(opts):
            base = self.binary

        docs = True
        if self.in_container(opts):
            docs = False

        if opts and 'docs' in opts:
            docs = opts['docs']

        docs_string = ""
        if not docs:
            docs_string = " --setopt=tsflags=nodocs"

        return base + docs_string


    def action(self, action, pkgs, options):
        return "{0} {1} {2}".format(
            self._base_command(options),
            action,
            " ".join(pkgs),
        ).strip()


    def install(self, pkgs, options=None):
        return self.action("install", pkgs, options)


    def reinstall(self, pkgs, options=None):
        return self.action("reinstall", pkgs, options)


    def remove(self, pkgs, options=None):
        return self.action("remove", pkgs, options)

    def update(self, pkgs, options=None):
        return self.action("update", pkgs, options)

    def update_all(self, options=None):
        return self.action("update", [], options);


    def cleancache(self, options=None):
        return self._base_command(options) + " clean all --enablerepo='*'"


class DnfPkgManager(YumPkgManager):
    binary = "dnf"


class CommandsConfig(object):
    interactive = False
    container = False


class Commands(object):
    def __init__(self, command_config, system_config):
        self.commands_config = command_config
        self.system_config = system_config

        pkginstid = system_config['package_installer']['name']
        if "yum" == pkginstid:
            self.pkginstaller = YumPkgManager(command_config)
        elif "dnf" == pkginstid:
            self.pkginstaller = DnfPkgManager(command_config)



from __future__ import print_function

import os, sys
import imp
import jinja2
from distgen.err import fatal
from distgen.pathmanager import PathManager
from distgen.config import load_config, merge_yaml
from distgen.project import AbstractProject
from distgen.commands import Commands
from distgen.multispec import Multispec, MultispecError


class Generator(object):
    project = None

    pm_cfg = None
    pm_tpl = None
    pm_spc = None


    def __init__(self):
        self.pm_cfg = PathManager(
            [os.path.join(sys.prefix, "share", "distgen", "distconf")],
            envvar="DG_DISTCONFDIR"
        )

        self.pm_tpl = PathManager(
            [os.path.join(sys.prefix, "share", "distgen", "templates")],
            envvar="DG_TPLDIR"
        )

        self.pm_spc = PathManager([])


    def load_project(self, project):
        self.project = self._load_project_from_dir(project)
        if not self.project:
            self.project = AbstractProject()
        self.project.directory = project

        def absolute_load(name):
            """
            In our templating system, we care about filenames specified by
            absolute path, which is not truth for default FileSystemLoader.
            """
            if name.startswith('/'):
                try:
                    with open(name, 'rb') as f:
                        return f.read().decode('utf-8')
                except:
                    pass
            raise jinja2.TemplateNotFound(name)

        def string_load(name):
            """
            Allow specifying a string instead of template to be able
            to return expanded config/specs/...
            """
            if name.startswith(('{{', '{%')):
                return name
            raise jinja2.TemplateNotFound(name)

        loader = jinja2.ChoiceLoader([
            jinja2.FileSystemLoader(self.pm_tpl.get_path()),
            jinja2.FunctionLoader(absolute_load),
            jinja2.FunctionLoader(string_load),
        ])

        self.project.tplgen = jinja2.Environment(
            loader=loader,
            keep_trailing_newline=True,
        )

        self.project.abstract_initialize()


    @staticmethod
    def _load_python_file(filename):
        """ load compiled python source """
        mod_name, file_ext = os.path.splitext(os.path.split(filename)[-1])
        if file_ext.lower() == '.py':
            py_mod = imp.load_source(mod_name, filename)
        elif file_ext.lower() == '.pyc':
            py_mod = imp.load_compiled(mod_name, filename)

        return py_mod


    def _load_obj_from_file(self, filename, objname):
        py_mod = self._load_python_file(filename)

        if hasattr(py_mod, objname):
            return getattr(py_mod, objname)
        else:
            return None


    def _load_obj_from_projdir(self, projectdir, objname):
        """ given project directory, load possibly existing project.py """
        project_file = projectdir + "/project.py"

        if os.path.isfile(project_file):
            return self._load_obj_from_file(project_file, objname)
        else:
            return None


    def _load_project_from_dir(self, projectdir):
        """ given project directory, load possibly existing project.py """
        projclass = self._load_obj_from_projdir(projectdir, "Project")
        if not projclass:
            return None
        return projclass()


    def load_config_from_project(self, directory):
        """
        read the project.py file for macros
        """
        config = self._load_obj_from_projdir(directory, 'config')
        if config:
            return config
        return {}


    @staticmethod
    def vars_fixed_point(config):
        """ substitute variables in paths """

        keys = config.keys()

        something_changed = True
        while something_changed:
            something_changed = False

            for i in keys:
                for j in keys:
                    if j == i:
                        continue
                    replaced = config[i].replace("$" + j, config[j])
                    if replaced != config[i]:
                        something_changed = True
                        config[i] = replaced


    def vars_fill_variables(self, config, sysconfig=None):
        if not 'macros' in config:
            return

        macros = config['macros']

        additional_macros = {}
        if sysconfig and 'macros' in sysconfig:
            additional_macros = sysconfig['macros']

        merged = merge_yaml(additional_macros, macros)
        if 'name' in config:
            merged['name'] = config['name']
        else:
            merged['name'] = 'unknown-pkg'
        self.vars_fixed_point(merged)

        config['macros'] = {x: merged[x] for x in macros.keys()}


    def render(self, specfiles, multispec, multispec_selectors, template,
               config, cmd_cfg, output=sys.stdout, confdirs=None,
               explicit_macros={}):
        """ render single template """
        config_path = [self.project.directory] + self.pm_cfg.get_path()
        sysconfig = load_config(config_path, config)

        if not confdirs:
            confdirs = []
        for i in confdirs + [self.project.directory]:
            additional_vars = self.load_config_from_project(i)
            self.vars_fill_variables(additional_vars, sysconfig)
            # filter only interresting variables
            interresting_parts = ['macros']
            additional_vars = {x: additional_vars[x] \
                    for x in interresting_parts if x in additional_vars}
            sysconfig = merge_yaml(sysconfig, additional_vars)

        self.project.abstract_setup_vars(sysconfig)

        self.project.inst_init(specfiles, template, sysconfig)

        projcfg = self.load_config_from_project(self.project.directory)
        if projcfg and 'name' in projcfg:
            sysconfig['name'] = projcfg['name']
        self.vars_fill_variables(sysconfig)

        explicit_macros = {'macros': explicit_macros}
        self.vars_fill_variables(explicit_macros, sysconfig)
        sysconfig = merge_yaml(sysconfig, explicit_macros)

        # NOTE: This is soo ugly, sorry for that, in future we need to modify
        # PyYAML to let us specify callbacks, somehow.  But for now, import
        # yaml right here to be able to add the constructors/representers
        # "locally".
        import yaml

        def _eval_node(loader, node):
            return str(eval(str(loader.construct_scalar(node)), {
                'project': self.project,
                'config': sysconfig,
                'macros': sysconfig['macros'],
            }))

        yaml.add_constructor(u'!eval', _eval_node)

        spec = {}
        for specfile in specfiles or []:
            specfd = self.pm_spc.open_file(
                specfile,
                [self.project.directory],
                fail=True,
            )
            if not specfd:
                fatal("Spec file {0} not found".format(specfile))

            try:
                specdata = yaml.load(specfd)
                spec = merge_yaml(spec, specdata)
            except yaml.YAMLError as exc:
                fatal("Error in spec file: {0}".format(exc))
        if multispec:
            try:
                mltspc = Multispec.from_path(self.project.directory, multispec)
                spec = merge_yaml(spec, mltspc.select_data(multispec_selectors, config))
            except yaml.YAMLError as exc:
                fatal("Error in multispec file: {0}".format(exc))
            except MultispecError as exc:
                fatal(str(exc))

        try:
            tpl = self.project.tplgen.get_template(template)
        except jinja2.exceptions.TemplateNotFound as err:
            fatal("Can not find template {0}".format(err))

        self.project.inst_finish(specfiles, template, sysconfig, spec)

        output.write(tpl.render(
            config=sysconfig,
            macros=sysconfig["macros"],
            m=sysconfig["macros"],
            container={'name': 'docker'},
            spec=spec,
            project=self.project,
            commands=Commands(cmd_cfg, sysconfig),
            env=os.environ,
        ))



import pytest

from distgen.commands import DnfPkgManager, YumPkgManager

class TestIndividualPkgManagers(object):
    def setup_method(self, method):
        class Config(object):
            def __init__(self, interactive, container=None):
                self.interactive = interactive
                self.container = container

        self.mgrs = {'yumi': YumPkgManager(Config(True)),
                     'yumni': YumPkgManager(Config(False)),
                     'dnfi': DnfPkgManager(Config(True)),
                     'dnfni': DnfPkgManager(Config(False)),
                     }

    @pytest.mark.parametrize('m, action, result', [
        ('yumi', 'install', 'yum install foo bar'),
        ('yumni', 'install', 'yum -y install foo bar'),
        ('yumi', 'reinstall', 'yum reinstall foo bar'),
        ('yumni', 'reinstall', 'yum -y reinstall foo bar'),
        ('yumi', 'remove', 'yum remove foo bar'),
        ('yumni', 'remove', 'yum -y remove foo bar'),
        ('yumi', 'update', 'yum update foo bar'),
        ('yumni', 'update', 'yum -y update foo bar'),
        ('dnfi', 'install', 'dnf install foo bar'),
        ('dnfni', 'install', 'dnf -y install foo bar'),
        ('dnfi', 'reinstall', 'dnf reinstall foo bar'),
        ('dnfni', 'reinstall', 'dnf -y reinstall foo bar'),
        ('dnfi', 'remove', 'dnf remove foo bar'),
        ('dnfni', 'remove', 'dnf -y remove foo bar'),
        ('dnfi', 'update', 'dnf update foo bar'),
        ('dnfni', 'update', 'dnf -y update foo bar'),
    ])
    def test_pkg_methods(self, m, action, result):
        assert getattr(self.mgrs[m], action)(['foo', 'bar']) == result

    @pytest.mark.parametrize('m, result', [
        ('yumi', 'yum update'),
        ('yumni', 'yum -y update'),
        ('dnfi', 'dnf update'),
        ('dnfni', 'dnf -y update'),
    ])
    def test_update_all(self, m, result):
        assert self.mgrs[m].update_all() == result

    @pytest.mark.parametrize('m, result', [
        ('yumi', 'yum clean all --enablerepo=\'*\''),
        ('yumni', 'yum -y clean all --enablerepo=\'*\''),
        ('dnfi', 'dnf clean all --enablerepo=\'*\''),
        ('dnfni', 'dnf -y clean all --enablerepo=\'*\''),
    ])
    def test_cleancache(self, m, result):
        assert self.mgrs[m].cleancache() == result



import os

import pytest
import six

from distgen.commands import CommandsConfig
from distgen.generator import Generator
from distgen.project import AbstractProject


here = os.path.dirname(__file__)
fixtures = os.path.join(here, 'fixtures', 'generator')
simple = os.path.join(fixtures, 'simple')


class TestGenerator(object):
    def setup_method(self, method):
        self.g = Generator()

    def test_load_project(self):
        self.g.load_project(simple)
        assert isinstance(self.g.project, AbstractProject)
        assert self.g.project.directory == simple
        assert hasattr(self.g.project, 'tplgen')

    @pytest.mark.parametrize('config, result', [
        ({'foo': 'bar'}, {'foo': 'bar'}),
        ({'foo': 'bar', 'baz': '$foo-asd'}, {'foo': 'bar', 'baz': 'bar-asd'}),
        ({'foo': 'bar', 'baz': '$foo-asd', 'x': 'x-$baz'},
         {'foo': 'bar', 'baz': 'bar-asd', 'x': 'x-bar-asd'}),
    ])
    def test_vars_fixed_point(self, config, result):
        self.g.vars_fixed_point(config)
        assert config == result

    @pytest.mark.parametrize('config, sysconfig, result', [
        ({'macros': {'x': '$y', 'y': 'z'}}, None, {'macros': {'x': 'z', 'y': 'z'}}),
        ({'macros': {'x': '$y'}}, {'macros': {'y': 'z'}}, {'macros': {'x': 'z'}}),
    ])
    def test_fill_variables(self, config, sysconfig, result):
        self.g.vars_fill_variables(config, sysconfig)
        assert config == result

    @pytest.mark.parametrize('template, result', [
        (os.path.join(simple, 'Dockerfile'), open(os.path.join(simple, 'expected_output')).read()),
        ('{{ config.os.id }}', 'fedora'),
    ])
    def test_render(self, template, result):
        # TODO: more test cases for rendering
        self.g.load_project(simple)
        out = six.StringIO()
        self.g.render(
            [os.path.join(simple, 'common.yaml')],
            os.path.join(simple, 'complex.yaml'),
            ['version=2.4', 'something_else=foo'],
            template,
            'fedora-26-x86_64.yaml',
            CommandsConfig(),
            out,
        )

        assert out.getvalue() == result



from setuptools import setup
from distgen.version import dg_version
from os import listdir, path

project = "distgen"
datadir = "share"
pkgdatadir = datadir + "/" + project
tpldir = pkgdatadir + "/templates"
distconfdir = pkgdatadir + "/distconf"

def dynamic_data_files():
    dynamic_list = []
    for dcdir in ["distconf", "distconf/lib"]:
        dcfiles = ["{0}/{1}".format(dcdir, f)
                   for f in listdir(dcdir)]
        dcfiles = [f for f in dcfiles if path.isfile(f)]
        dynamic_list.append((pkgdatadir + "/" + dcdir, dcfiles))

    return dynamic_list

def get_requirements():
    with open('requirements.txt') as f:
        return f.read().splitlines()

setup(
    name='distgen',
    version=dg_version,
    description='Templating system/generator for distributions',
    author='praiskup@redhat.com',
    license='GPLv2+',
    url='https://github.com/devexp-db/distgen',
    platforms=['any'],
    packages=['distgen'],
    data_files=[
        (tpldir + '/container/docker', [
            'templates/container/docker/parts.tpl',
        ]),
        (tpldir, [
            'templates/docker.tpl',
            'templates/makefile-macros.tpl',
            'templates/README',
            'templates/general.tpl',
        ]),
    ] + dynamic_data_files(),
    scripts=['bin/dg'],
    install_requires=get_requirements(),
)



class AbstractProject():
    tplgen = None

    name = "unknown-project"

    maintainer = "unknown <unknown@unknown.com>"


    def inst_init(self, specfiles, template, sysconfig):
        """
        Executed before the project.py/spec files/template is loaded and
        before all the dynamic stuff and specification is calculated.
        Now is still time to dynamically change the list of specfiles or
        adjust the system configuration. You can define a variable as an
        attribute of this project:

          self.variable = "42"

        which can be later utilized in a template like this:

          {{ project.variable }}
        """
        pass


    def inst_finish(self, specfiles, template, sysconfig, spec):
        """
        Executed after the project.py/spec files/template is loaded, and
        the specification (spec) calculated (== instantiated).  This is
        the last chance to dynamically change sysconfig or spec.
        """
        pass


    def __init__(self):
        """ Never overwrite constructor please! """
        pass

    def abstract_initialize(self):
        """
        Never overwrite this function, please.  Its currently just a
        placeholder called right after Project was instantiated and the
        right paths to templates/specs was set.
        """
        pass

    def abstract_setup_vars(self, config):
        """
        Never overwrite this function, please.
        """

        # Be careful here with changes.  Any change will survive to the next
        # call of render() method (its effect won't disappear for next
        # template).
        pass
